# Summary statistics

from cupyx.scipy.stats._distributions import entropy  # NOQA
from cupyx.scipy.stats._stats import trim_mean  # NOQA


# Other statistical functionality

from cupyx.scipy.stats._morestats import boxcox_llf  # NOQA
from cupyx.scipy.stats._stats_py import zmap  # NOQA
from cupyx.scipy.stats._stats_py import zscore  # NOQA
