// This file is a stub header file of cuda for Read the Docs.

#ifndef INCLUDE_GUARD_STUB_CUPY_CUDA_H
#define INCLUDE_GUARD_STUB_CUPY_CUDA_H

#include "cupy_cuda_common.h"  // for CUDA_VERSION

extern "C" {

// Occupancy
typedef size_t (*CUoccupancyB2DSize)(int);

} // extern "C"

#endif // #ifndef INCLUDE_GUARD_STUB_CUPY_CUDA_H
